#!/usr/bin/env python
# encoding: utf-8
"""
Mòdul que implementa la classe que representa una execució de l'algoritme PSO i les particules necessaries per fer-ho.

Created by Joan T. Matamalas Llodrà on 2011-11-11.
Copyright (c) 2011 . All rights reserved.
"""
import random
import time
import math
from algorithms.ec import Ec
from functions import *

class Particle(object):
    """Métode creador de la classe partícula"""
    def __init__(self, rangeA, rangeB, dimensions, vMax):
        super(Particle, self).__init__()
        self.currentPos = [random.uniform(rangeA, rangeB) for n in range(dimensions)] #Posició actual
        self.lbest = {'pos':[],'score':0} #Millor posició aconseguida per la particula
        self.nbest = {'pos':[],'score':0} #Millor posició aconseguida per el conjunt de veïns
        self.neighborhood = [] #Conjunt de veïns
        self.velocity = [random.uniform(-vMax, vMax) for n in range(dimensions)] #Velocitat inicial
        
    def setLbest(self, lb, lbScore):
        """Mètode que actualitza l'informació de la millor posició aconseguida fins el moment
        i transmet aquesta informació als veïns per tal que, si procedeix, actualitzin l'informació
        de millor posició del conjunt de veïns"""
        self.lbest['pos'] = lb[:]
        self.lbest['score'] = lbScore
        if lbScore < self.nbest['score']:
            self.nbest['pos'] = lb[:]
            self.nbest['score'] = lbScore
            if self.neighborhood[0].nbest['score'] > lbScore:
                self.neighborhood[0].nbest['pos'] = lb[:]
                self.neighborhood[0].nbest['score'] = lbScore
            if self.neighborhood[2].nbest['score'] > lbScore:
                self.neighborhood[2].nbest['pos'] = lb[:]
                self.neighborhood[2].nbest['score'] = lbScore
        
    def findNBest(self):
        """Métode que comprova la millor solució trobada fins el moment per el conjunt de veïns"""
        for n in self.neighborhood:
            if n.lbest['score'] < self.nbest['score']:
                self.nbest['pos'] = n.lbest['pos'][:]
                self.nbest['score'] = n.lbest['score']

class Pso(Ec):
    """Mètode creador de la classe PSO"""
    def __init__(self, function, seed, iterations, threshold=1e-10, nPop = 50, chi=0.73, phiP = 2.05, phiG = 2.05):
        super(Pso, self).__init__(function, seed, iterations, threshold, nPop = 50)
        self.population = self.initPopulation() #Iniciam la població de particules
        self.chi = chi #Constricció simple de la velocitat
        self.phiP = 2.05 #Influència de la millor posició aconseguida per la particula
        self.phiG = 2.05 #Influència de la millor posició aconseguida pel conjunt de veïns
        
    def initPopulation(self):
        """Mètode que inicia una població de particules"""
        #Iniciam el generador pseudoaleatori amb una determinada llavor 
        random.seed(self.seed)
        vMax = (self.function.rangeB-self.function.rangeA)*0.15 #Velocitat màxima inicial de les partícules
        pop = [Particle(self.function.rangeA, self.function.rangeB, self.function.dimensions, vMax) for x in range(self.nPop)]
        
        # Iniciam neighborhood
        for i in range(len(pop)):
            pop[i].neighborhood = [pop[(i-1)%self.nPop],pop[i],pop[(i+1)%self.nPop]]
            
        # Determinam les millor posicions coneguedes
        for p in pop:
            p.lbest['pos'] = p.currentPos[:]
            p.nbest['pos'] = p.currentPos[:]
            evalCurrent = self.function.evalf(p.currentPos)
            p.lbest['score'] = evalCurrent
            p.nbest['score'] = evalCurrent

        for p in pop:
            p.findNBest()
        
        return pop
        
    def run(self):
        """Mètode que executa l'algoritme PSO"""
        self.startTime = time.time()
       
        #Emmagatzemam la millor posició aconseguida
        gbest = self.population[0].currentPos[:]
        bestValue = self.function.evalf(self.population[0].currentPos)
    
        for p in self.population:
            if(self.function.evalf(p.currentPos) < bestValue):
                gbest = p.currentPos
                bestValue = self.function.evalf(p.currentPos)
                
        self.bestInv = gbest
        self.bestFitness = bestValue
        
        for iteration in range(self.iterations):
            for p in self.population:
                # Actualitzam la velocitat
                for d in range(self.function.dimensions):
                    randP = random.uniform(0,self.phiP)
                    randG = random.uniform(0,self.phiG)
                    p.velocity[d] = self.chi*((p.velocity[d]) + randP*(p.lbest['pos'][d]-p.currentPos[d]) + randG*(p.nbest['pos'][d]-p.currentPos[d]))

                    p.currentPos[d] = p.currentPos[d] + p.velocity[d]
                    
                    #Limitam la posició al rang de cerca
                    if p.currentPos[d] > self.function.rangeB:
                        p.currentPos[d] = self.function.rangeB
                    elif p.currentPos[d] < self.function.rangeA:
                        p.currentPos[d] = self.function.rangeA
                    
                #Evaluam la nova posició
                posFitness = self.function.evalf(p.currentPos)
            
                #Actualitzam si procedeix les dades de millor posició aconseguida.
                if posFitness < p.lbest['score']:
                    p.setLbest(p.currentPos[:],posFitness)
                    if posFitness < bestValue:
                        bestValue = posFitness
                        gbest = p.currentPos[:]
        
            #Actualitzam els resultats
            if bestValue < self.bestFitness:
                self.bestFitness = bestValue
                self.bestIndv = gbest[:]
            self.iterationsUntilGoal += 1
            
            #Comprovam si s'ha arribat a l'òptim
            if bestValue < self.function.minimum+self.threshold:
                self.bestFitness = self.function.minimum
                self.isGoalFounded = True
                break
            
        self.stopTime = time.time()