#!/usr/bin/env python
# encoding: utf-8
"""
Aquest modul és el que controla l'execució de l'experiment i retornar els resultats

Created by Joan T. Matamalas Llodrà on 2011-11-15.
Copyright (c) 2011 . All rights reserved.
"""

import sys
import time
import getopt
import numpy as np
import shelve

from terminalProgressBar.progressBar import ProgressBar

from algorithms.ga import Ga
from algorithms.de import De
from algorithms.pso import Pso
from algorithms.abc import Abc
from functions import *
from result import *
from scipy import stats
from multiprocessing import Process, Queue, Manager, cpu_count, Value
from operator import itemgetter, attrgetter

help_message = '''

Aquestes són les opcions disponibles per iniciar l\'execució de l\'experiment:

    -v\\--verbose           Mostra els estadistics de les diferents execucions per pantalla
    -i\\--iterations=[0-9]* Nombre màxims d'iteracions per execució. Per defecte 50.000
    -e\\--numExec=[0-9]*    Nombre d'execucions per funció i algoritme. Per defecte 31
    -s\\--seed=[0-9]*       Llavor del generador pseudoaleatori. Per defecte 42
    -i\\--input={filename}  Permet seleccionar un fitxer d'entrada amb dades sobre una execució anterior.
    -o\\--output={filename} Crea un fitxer amb les dades dels objectes de l'execució
    -l\\--latex             Crea fitxers amb els resultats en format LaTeX
'''


class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

def get_max_width(table, index):
    """Get the maximum width of the given column index
    """
    
    return max([len(str(row[index])) for row in table])

def pprint_table(out, table):
    """Prints out a table of data, padded for alignment
    
    @param out: Output stream ("file-like object")
    @param table: The table to print. A list of lists. Each row must have the same
    number of columns.
    
    """

    col_paddings = []
    
    for i in range(len(table[0])):
        col_paddings.append(get_max_width(table, i))

    for row in table:
        # left col
        print >> out, row[0].ljust(col_paddings[0] + 1),
        # rest of the cols
        for i in range(1, len(row)):
            col = str(row[i]).rjust(col_paddings[i] + 2)
            print >> out, col,
        print >> out

def formatTime(seconds):
    """La cadena de temps en format hores, minuts, segons"""
    out = ''
    seconds = int(seconds)
    if seconds//3600 > 0:
        hours = seconds//3600
        seconds -= hours*3600
        minutes = seconds//60
        seconds -= minutes*60
        out += '%d hores %d minuts %d segons'%(hours,minutes, seconds,)
    elif seconds//60 > 0:
        minutes = seconds//60
        seconds -= minutes*60
        out += '%d minuts %d segons' %(minutes, seconds)
    else:
        out += '%d segons' %seconds
    return out

def exeProcess(algo, function, s, ite, mutex, semaphore, results):
    """Mètode que executa un algroitme i retorna el resulta. Aquést mètode s'executa com a procés independent"""
    #Inicia l'algoritme
    result = algo(function, seed = s,iterations = ite)
    #Inicia l'execució
    result.run()
    
    #Adquireix el testimoni per tal d'evitar un ús mutu de la llista compartida
    mutex.acquire()
    
    #Afegeix el resultat a la llista compartida
    results.append(result)
    
    #Allibera el testimoni
    mutex.release()
    
    #Finalitza l'execució alliberant un slot d'execució perque pugui ser utilitzat per una altre procés.
    semaphore.release()

def exe(algo, function, s, ite, nExec):
    """Mètode que executa """
    #Inicia el controlador de l'execució multiprocés
    manager = Manager()
    
    #Inicia un testimoni compartit
    mutex = manager.Lock()
    
    #Inicia un semafor compartit, amb tants slots d'execució com nuclis de procés té la CPU
    semaphore = manager.BoundedSemaphore(cpu_count())
    
    #Inicia una llista compartida entre els processos
    results = manager.list()    
    
    #Crea nExec processos d'un determinat algoritme i funció
    processes = [Process(target=exeProcess, args=(algo, function, s+i, ite, mutex, semaphore, results)) for i in range(nExec)]
    
    #Mentre hi ha algún procés pendent, assigna els slots depenguent de la disponibilitat
    for process in processes:
        #Quan és possible adquireix un slot
        semaphore.acquire()
        #Inicia el procés
        process.start()
    
    #Esperam a acabar totes les execucions
    while len(results) < nExec:
        pass
        
    #retorna els resultats
    return results
    
def compare(alg1,alg2,resultsDic,significance):
    """Mètode que realitza la comparativa entre 2 algoritmes, basant-se en les funcions disponibles i el nivell de significació"""
    #Funcions a comparar
    functions = ['sphere','step','sumSquares','colville','zakharov','schwefel2p22','himmelblau','rastrigin','schwefel','michalewicz5','schaffer','shubert','griewank','ackley','shekel10']
    #Resultats de les comparacions
    comparations = []
    
    #Per a cada funció, fa un t_test entre els dos algoritmes
    for f in functions:
        test = stats.ttest_ind(resultsDic[alg1][f]['stats'].best, resultsDic[alg2][f]['stats'].best)
        comparations.append(comparationResult(alg1, alg2, f, test[0], test[1]))
    
    #Comparatives ordenades de major a menor tenint en compte el p_value
    comparations.sort(key=attrgetter('p_value'),reverse=True)
    
    #Realitza el contrast d'hipotesi per a cada funció
    for i in range(len(comparations)):
        comparations[i].rank = i + 1
        #Modificació de la significació deguda al mètode de Bonferroni
        comparations[i].significance = significance/comparations[i].rank
        #Contrast d'hipotesi
        comparations[i].hValidation()
    
    #Retorna els resultats
    return comparations

def formatTable(table):
    """Mètode que formata els nombres en coma flotant de una determinada taula"""
    for i in range(len(table)):
        for j in range(len(table[0])):
            if isinstance(table[i][j], float):
                fnumber = round(table[i][j], 10)
                table[i][j] = '%g'%fnumber
    
    return table
    
def latexOutput(table, filename):
    """Crea dos arxius amb els resultats tabulats en format LaTeX"""
    if not os.path.exists('resultatsLatex'):
        os.makedirs('resultatsLatex')
        
    rowString = ['&'.join([str(c) for c in row]) for row in table]
    outputL = '\\\\\n'.join(rowString)+'\\\\'
    f = open('resultatsLatex/'+filename+'.tex', 'w')
    f.write(outputL)
    f.close()
    
def resultsTable(resultsDic, functions):
    """Mètode que crea una taula amb els estadistics de les execucions"""
    table = [['Funció','Mínim','','GA','DE','PSO','ABC']]
    for f in functions:
        row = [f.name, f.minimum,'Fitness mean',resultsDic['GA'][f.name]['stats'].avgFitness,resultsDic['DE'][f.name]['stats'].avgFitness,resultsDic['PSO'][f.name]['stats'].avgFitness,resultsDic['ABC'][f.name]['stats'].avgFitness]
        table.append(row[:])
        row = ['','','Fitness stdDev',resultsDic['GA'][f.name]['stats'].stdFitness,resultsDic['DE'][f.name]['stats'].stdFitness,resultsDic['PSO'][f.name]['stats'].stdFitness,resultsDic['ABC'][f.name]['stats'].stdFitness]
        table.append(row[:])
        row = ['','','Fitness SEM',resultsDic['GA'][f.name]['stats'].semFitness,resultsDic['DE'][f.name]['stats'].semFitness,resultsDic['PSO'][f.name]['stats'].semFitness,resultsDic['ABC'][f.name]['stats'].semFitness]
        table.append(row[:])
        row = ['','','Min Fitness',resultsDic['GA'][f.name]['stats'].minFitness,resultsDic['DE'][f.name]['stats'].minFitness,resultsDic['PSO'][f.name]['stats'].minFitness,resultsDic['ABC'][f.name]['stats'].minFitness]
        table.append(row[:])
        row = ['','','Max Fitness',resultsDic['GA'][f.name]['stats'].maxFitness,resultsDic['DE'][f.name]['stats'].maxFitness,resultsDic['PSO'][f.name]['stats'].maxFitness,resultsDic['ABC'][f.name]['stats'].maxFitness]
        table.append(row[:])
        
        row = ['','','Success executions percent',resultsDic['GA'][f.name]['stats'].successExec,resultsDic['DE'][f.name]['stats'].successExec,resultsDic['PSO'][f.name]['stats'].successExec,resultsDic['ABC'][f.name]['stats'].successExec]
        table.append(row[:])
            
        row = ['','','Iterations mean',resultsDic['GA'][f.name]['stats'].avgIterations,resultsDic['DE'][f.name]['stats'].avgIterations,resultsDic['PSO'][f.name]['stats'].avgIterations,resultsDic['ABC'][f.name]['stats'].avgIterations]
        table.append(row[:])
        row = ['','','Iterations stdDev',resultsDic['GA'][f.name]['stats'].stdIterations,resultsDic['DE'][f.name]['stats'].stdIterations,resultsDic['PSO'][f.name]['stats'].stdIterations,resultsDic['ABC'][f.name]['stats'].stdIterations]
        table.append(row[:])
        row = ['','','Iterations SEM',resultsDic['GA'][f.name]['stats'].semIterations,resultsDic['DE'][f.name]['stats'].semIterations,resultsDic['PSO'][f.name]['stats'].semIterations,resultsDic['ABC'][f.name]['stats'].semIterations]
        table.append(row[:])
        row = ['','','Min Iterations',resultsDic['GA'][f.name]['stats'].minIterations,resultsDic['DE'][f.name]['stats'].minIterations,resultsDic['PSO'][f.name]['stats'].minIterations,resultsDic['ABC'][f.name]['stats'].minIterations]
        table.append(row[:])
        row = ['','','Max Iterations',resultsDic['GA'][f.name]['stats'].maxIterations,resultsDic['DE'][f.name]['stats'].maxIterations,resultsDic['PSO'][f.name]['stats'].maxIterations,resultsDic['ABC'][f.name]['stats'].maxIterations]
        table.append(row[:])

        row = ['','','Duration mean',resultsDic['GA'][f.name]['stats'].avgTime, resultsDic['DE'][f.name]['stats'].avgTime, resultsDic['PSO'][f.name]['stats'].avgTime,resultsDic['ABC'][f.name]['stats'].avgTime]
        table.append(row[:])
        row = ['','','Duration stdDev',resultsDic['GA'][f.name]['stats'].stdTime,resultsDic['DE'][f.name]['stats'].stdTime,resultsDic['PSO'][f.name]['stats'].stdTime,resultsDic['ABC'][f.name]['stats'].stdTime]
        table.append(row[:])
        row = ['','','Duration SEM',resultsDic['GA'][f.name]['stats'].semTime,resultsDic['DE'][f.name]['stats'].semTime,resultsDic['PSO'][f.name]['stats'].semTime,resultsDic['ABC'][f.name]['stats'].semTime]
        table.append(row[:])
        row = ['','','Min Duration',resultsDic['GA'][f.name]['stats'].minTime,resultsDic['DE'][f.name]['stats'].minTime,resultsDic['PSO'][f.name]['stats'].minTime,resultsDic['ABC'][f.name]['stats'].minTime]
        table.append(row[:])
        row = ['','','Max Duration',resultsDic['GA'][f.name]['stats'].maxTime,resultsDic['DE'][f.name]['stats'].maxTime,resultsDic['PSO'][f.name]['stats'].maxTime,resultsDic['ABC'][f.name]['stats'].maxTime]
        table.append(row[:])
        row = ['','','','','','','']
        table.append(row[:])
            
    return formatTable(table)
        
def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        try:
            opts, args = getopt.getopt(argv[1:], "hln:o:i:e:s:v", ["help","latex", "iterations=","output=", "input=", "numExec=",'seed','verbose'])
        except getopt.error as msg:
            raise Usage(msg)
        
        nExec = 31 #Nombre per defecte d'execucions per a cada algoritme i funció
        seed = 42 #Valor per defecte de la llavor del generador pseudoaleatori
        iterations = 50000 #Nombre per defecte d'iteracions màximes per execució de cada algoritme
        isOutputSelected = False #S'ha seleccionat un fitxer de sortida?
        isInputSelected = False #S'ha seleccionat un fitxer d'entrada?
        latexOut = False #S'han de crear fitxers amb els resultats en format latex?
        verbose = False #S'han de mostrar els resultats per pantalla?
        ouputData = 'output.data' #Nom de l'arxiu de dades per defecte
        inputData = 'output.data' #Nom de l'arxiu de dades per defecte
        
        # option processing
        for option, value in opts:
            if option in ('-v','--verbose'):
                verbose = True
            if option in ("-l", "--latex"):
                latexOut = True
            if option in ("-h", "--help"):
                raise Usage(help_message)
            if option in ("-o", "--output"):
                isOutputSelected = True
                outputData = value #Adreça del fitxer de sortida
            if option in ('-i', '--input'):
                isInputSelected = True
                inputData = value #Adreça del fitxer d'entrada
            if option in ("-n", "--iterations"):
                iterations = int(value)
            if option in ('-e','--numExec'):
                nExec = int(value)
            if option in ('-s','--seed'):
                seed = int(seed)
                

        #Iniciam les confiuracions de les funcions a optimitzar
        functions = [function('sphere', sphere, 30, -100, 100, 0),
                    function('step',step,30,-100,100,0),
                    function('sumSquares',sumSquares,30,-10,10,0),
                    function('colville',colville,4,-10,10,0),
                    function('zakharov',zakharov,10,-5,10,0),
                    function('schwefel2p22',schwefel2p22,30,-10,10,0),
                    function('himmelblau',himmelblau,2,-10,10,0),                        
                    function('rastrigin',rastrigin,30,-5.12,5.12,0),
                    function('schwefel',schwefel,30,-500,500,-12569.4866181730),
                    function('michalewicz5',michalewicz5,5,0,math.pi,-4.6876581790),
                    function('schaffer',schaffer,2,-100,100,0),
                    function('shubert',shubert,2,-10,10, -186.7309088310),
                    function('griewank',griewank,30,-600,600,0),
                    function('ackley',ackley,30,-32,32,0),
                    function('shekel10',shekel10,4,0,10,-10.5364098167)]

        if not isInputSelected:
            #Si no hi ha fitxer d'entrada seleccionat realitzam els calculs
            
            #Algoritmes a executar
            alg = {'GA':Ga, 'DE':De, 'ABC':Abc, 'PSO':Pso}
        
            #Diccionari amb els resultats de les execucions
            resultsDic = {}
            
            for k in alg.keys():
                resultsDic[k] = {'sphere':{'results':[]},
                                    'step':{'results':[]},
                                    'sumSquares':{'results':[]},
                                    'colville':{'results':[]},
                                    'zakharov':{'results':[]},
                                    'schwefel2p22':{'results':[]},
                                    'himmelblau':{'results':[]},
                                    'rastrigin':{'results':[]},
                                    'michalewicz5':{'results':[]},
                                    'schwefel':{'results':[]},
                                    'schaffer':{'results':[]},
                                    'shubert':{'results':[]},
                                    'griewank':{'results':[]},
                                    'ackley':{'results':[]},
                                    'shekel10':{'results':[]}}
        
            #Iniciam el comptador de temps
            startTime = time.time()
        
            print('\nCOMPARATIVA ABC vs GA, DE, PSO\n––––––—–––––––––––––––––––––––')
            
            #Iniciam la barra de progress
            pb = ProgressBar('blue',width=40, block='▣', empty='□')
            progress = 0
        
            #Realitzam les execucions actualitzant la barra de progrés amb l'informació sobre el percenatge d'execució completat
            for f in functions:
                for a in alg:
                    pb.render(int(progress),'%s\nOptimitzant funció: %s\nAlgortime utilitzat: %s'%(formatTime(time.time()-startTime),f.name,a,))
                    resultsDic[a][f.name]['results'] = exe(alg[a], f, seed, iterations, nExec)
                    resultsDic[a][f.name]['stats'] = statsResults(resultsDic[a][f.name]['results'])
                    progress += 100./(len(functions)*len(alg))
            pb.render(100,'Execucions completades')
            
            print('Temps utilitzat: %s'%formatTime(time.time()-startTime))
            print('Calculant resultats...\n')
        
            if isOutputSelected:
                #Si s'ha seleccionat un fitxer de sortida, desam els estadístics de l'execució.
                resFile = shelve.open(outputData)
                saveResults = {}
                for alg in resultsDic.keys():
                    saveResults[alg] = {}
                    for f in resultsDic[alg].keys():
                        saveResults[alg][f] = {}
                        saveResults[alg][f]['stats'] = resultsDic[alg][f]['stats']   
                
                outputshelve = shelve.open(outputData)
                outputshelve['results'] = saveResults
                outputshelve.close()                     
                
        else:
            #Si s'ha proporcionat una fitxer amb dades el carrega i mostra els resultats per pantalla
            inputshelve = shelve.open(inputData)
            resultsDic = inputshelve['results']
            inputshelve.close()
        
        table = resultsTable(resultsDic, functions)
        
        #Mostra els resultats de les execucions per pantalli si s'ha especificat a l'hora d'iniciar l'execucio 
        if verbose:
            pprint_table(sys.stdout,table)
        
        #Si s'ha marcat l'opció, crea un fitxer amb els resultats amb format LaTeX
        if latexOut:
            latexOutput(table,'resultatsExecucions')
        
        #Algoritmes amb els que comparam els ABC
        vs = ['GA','DE','PSO']
        
        #Mostra els resultats de la comparativa per pantalla
        for n in vs:
            print('\nCOMPARATIVA ABC vs %s:\n——————————————————————'%n)
            comparations = compare('ABC',n,resultsDic,0.05)
            table = [['Funció','t-value','p-value','Ranquing','Significancia','Millor algoritme']]
            for cf in comparations:
                    table.append([cf.function,cf.t_value,cf.p_value,cf.rank,cf.significance,cf.betterAlg()])
                    
            table = formatTable(table)
            
            #Si s'ha marcat l'opció, crea un fitxer amb els resultats amb format LaTeX
            if latexOut:
                latexOutput(table,'comparativaABCvs'+n)
               
            pprint_table(sys.stdout,table)
            print('\n')
            
    except Usage as err:
        print >> sys.stderr, sys.argv[0].split("/")[-1] + ": " + str(err.msg)
        print >> sys.stderr, "\t for help use --help"
        return 2


if __name__ == "__main__":
    sys.exit(main())
