#!/usr/bin/env python
# encoding: utf-8
"""
Aquest mòdul conté les classes que representen els estadístics de les execucions, statsResults,
i les comparatives realitzades, comparationResult

Created by Joan T. Matamalas Llodrà on 2011-11-15.
Copyright (c) 2011 . All rights reserved.
"""
import numpy as np
from scipy import stats

class statsResults(object):
    """Classe que representa els estadistics per un algoritme i funció determinats"""
    def __init__(self, results):
        super(statsResults, self).__init__()
        bestList = [result.bestFitness for result in results] #Llista amb el millor resultat de cada execució
        self.algorithm = results[0].__class__.__name__ #Nom de l'algoritme
        self.function = results[0].function.evalf.__name__ #Nom de la funció
        self.best = np.array(bestList) #Generam un numpy array amb els millor resultat de cada execució
        self.avgFitness = np.mean(self.best) #MBF
        self.stdFitness = np.std(self.best) #desviació estandard de l'MBF
        self.semFitness = stats.sem(self.best) #error estandard de l'MBF
        self.maxFitness = self.best.max() #Pitjor solució trobada al llarg de les execucions
        self.minFitness = min(self.best) #Millor solució trobada al llarg de les execucions
        timesList = [r.getTotalTime() for r in results if r.isGoalFounded] #Llista amb el temps utilitzat per a cada execucio finalitzada amb èxit
        self.successExec =  '%.2f%%'%(100*(len(timesList)/float(len(results)))) #Nombre d'execucions finalitzades amb èxit
        self.times = np.array(timesList) #Numpy array amb els temps
        self.avgTime = np.mean(self.times) if timesList else '-' #Temps mitja de les execucions concloses amb èxit
        self.stdTime = np.std(self.times) if timesList else '-' #Desviació estandard del temps
        self.semTime = stats.sem(self.times) if timesList else '-' #Error estandard del temps
        self.maxTime = max(self.times) if timesList else '-' #Major temps consumit
        self.minTime = min(self.times) if timesList else '-' #Menor temps consumit
        iterationsList = [r.iterationsUntilGoal for r in results if r.isGoalFounded] #Nombre d'iteracions necessaries per arribar a una solució exitosa.
        self.iterations = np.array(iterationsList) #Numpy array amb el nombre d'iteracions
        self.avgIterations = np.mean(self.iterations) if timesList else '-' #Nombre mitjà d'iteracions necessaries per arribar a una solució exitosa.
        self.stdIterations = np.std(self.iterations) if timesList else '-' #Desviació estàndard del nombre d'iteracions
        self.semIterations = stats.sem(self.times) if timesList else '-' #Error estàndard del nombre d'iteracions
        self.maxIterations = max(self.iterations) if timesList else '-' #Màxim nombre d'iteracions necessaries
        self.minIterations = min(self.iterations) if timesList else '-' #Mínim nombre d'iteracions necessaries
        
        
    def __str__(self):
        #Retorna els resultats en format string
        strRes = self.algorithm+' '+self.function+' results\n'
        strRes += 'Mean:\t'+str(self.mean)+'\n'
        strRes += 'STD:\t'+str(self.std)+'\n'
        strRes += 'SEM:\t'+str(self.sem)+'\n'
        return strRes
        
class comparationResult(object):
    """Classe que representa els resultats d'una comparativa"""
    def __init__(self, alg1, alg2, function, t_value, p_value):
        super(comparationResult, self).__init__()
        self.alg1 = alg1 #Primer algoritme
        self.alg2 = alg2 #Segon algoritme
        self.function = function #Funció d'avaluació
        self.t_value = t_value
        self.p_value = p_value
        self.rank = 0 #Ranking. Utilitzat amb conjunció a la correcció de Bonferroni
        self.significance = 0.0 #Nivell de significancià
        self.acceptNull = True #Acceptam hipòtesi nul·la?
        
    def hValidation(self):
        """Determina si acceptam o no la hipòtesi nul·la"""
        if self.p_value < self.significance:
            self.acceptNull = False
    
    def betterAlg(self):
        """Determina quin és el millor algoritme depenguent del seu t_value"""
        if self.acceptNull:
            return '-'
        elif self.t_value < 0:
            return self.alg1
        else:
            return self.alg2